/*
 * =================================================================
 * GbI2C.bas - GameBaby I2C device driver
 * =================================================================
 */
#include <propeller.h>
#include "GbI2c.h"

/*
 * =================================================================
 * PASM command interface
 * =================================================================
 */

#define GbI2C_CMD_ADC_READ          1
#define GbI2C_CMD_ACCEL_READ        2
#define GbI2C_CMD_ACCEL_WRITE       3
#define GbI2C_CMD_EEPROM_RDBUF      4
#define GbI2C_CMD_EEPROM_WRBUF      5
#define GbI2C_CMD_RTC_READ          6
#define GbI2C_CMD_RTC_WRITE         7

#define GbI2C_CMDMASK               0x7     // one of 7 states

#define GbI2C_scl_pin               28      // always 28.
#define GbI2C_sda_pin               29      // always 29.

#define GbI2C_ACCEL_ID              0x98    // Accelerometer device ID
#define GbI2C_ADC_ID                0x9A    // ADC device ID
#define GbI2C_EEPROM_ID             0xA0    // EEPROM device ID
#define GbI2C_RTC_ID                0xD0    // RTC device ID

#define GbI2C_EE_PAGELEN            128     // AT24C512 EEPROM PAGE LENGTH
#define GbI2C_RBIT                  1       // devid byte read bit

#define GbI2C_ADC_DELAY             1
#define GbI2C_ACCEL_DELAY           1
#define GbI2C_EEPROM_DELAY          1
#define GbI2C_RTC_DELAY             10

/*
=================================================================
ACCELEROMETER REGISTERS
=================================================================
Address Name Definition         Bit 7   Bit 6   Bit 5   Bit 4   Bit 3   Bit 2   Bit 1   Bit 0
$00 XOUT 6-bit output X     -       Alert   XOUT[5] XOUT[4] XOUT[3] XOUT[2] XOUT[1] XOUT[0]
$01 YOUT 6-bit output Y     -       Alert   YOUT[5] YOUT[4] YOUT[3] YOUT[2] YOUT[1] YOUT[0]
$02 ZOUT 6-bit output Z     -       Alert   ZOUT[5] ZOUT[4] ZOUT[3] ZOUT[2] ZOUT[1] ZOUT[0]
$03 TILT Tilt Status        Shake   Alert   Tap     PoLa[2] PoLa[1] PoLa[0] BaFro[1] BaFro[0]
$04 SRST Sampling Rate Stat 0       0       0       0       0       0       AWSRS   AMSRS
$05 SPCNT Sleep Count       SC[7]   SC[6]   SC[5]   SC[4]   SC[3]   SC[2]   SC[1]   SC[0]
$06 INTSU Interrupt Setup   SHINTX  SHINTY  SHINTZ  GINT    ASINT   PDINT   PLINT   FBINT
$07 MODE Mode               IAH     IPP     SCPS    ASE     AWE     TON     -       MODE
$08 SR Filter               FILT[2] FILT[1] FILT[0] AWSR[1] AWSR[0] AMSR[2] AMSR[1] AMSR[0]
$09 PDET Tap Detection      ZDA     YDA     XDA     PDTH[4] PDTH[3] PDTH[2] PDTH[1] PDTH[0]
$0A PD Tap Debounce Count   PD[7]   PD[6]   PD[5]   PD[4]   PD[3]   PD[2]   PD[1]   PD[0]
=================================================================
*/

#define GbI2C_ACCEL_XOUT    0
#define GbI2C_ACCEL_YOUT    1
#define GbI2C_ACCEL_ZOUT    2
#define GbI2C_ACCEL_TILT    3
#define GbI2C_ACCEL_SRST    4
#define GbI2C_ACCEL_SLPCNT  5
#define GbI2C_ACCEL_INTSU   6
#define GbI2C_ACCEL_MODE    7
#define GbI2C_ACCEL_SRFLT   8
#define GbI2C_ACCEL_PDET    9
#define GbI2C_ACCEL_PDCNT   10

/*
=================================================================
RTC REGISTERS
=================================================================
*/

#define GbI2C_RTC_SECOND    0
#define GbI2C_RTC_MINUTE    1
#define GbI2C_RTC_HOUR      2
#define GbI2C_RTC_DAY       3
#define GbI2C_RTC_DATE      4
#define GbI2C_RTC_MONTH     5
#define GbI2C_RTC_YEAR      6
#define GbI2C_RTC_CONTROL   7

/*
 * =================================================================
 * Button interface
 * =================================================================
 */
#define GbButtons_Up    1
#define GbButtons_Down  2
#define GbButtons_Left  4
#define GbButtons_Right 8

/*
=================================================================
Misc constants
=================================================================
*/

#define GbI2C_SMOOTHER    40
#define GbI2C_eeBufferLen 0x20

/*
=================================================================
eeprom buffer
=================================================================
*/
HUBDATA unsigned char GbI2C_eeBuffer[GbI2C_eeBufferLen];

/*
=================================================================
PASM Command interface
=================================================================
*/

HUBDATA unsigned char GbI2C_accelbuf[16] ; // 16 bytes
HUBDATA unsigned char GbI2C_rtcbuf[48]   ; // 64 bytes

int GbI2C_cog            ; // our cog number

HUBDATA int GbI2C_command; // command startup with scl & sda pins
HUBDATA int GbI2C_buttons; // buttons pointer
HUBDATA int GbI2C_device ; // default device number
HUBDATA int GbI2C_address; // register address
HUBDATA int GbI2C_count  ; // operation data length
HUBDATA int *GbI2C_bufptr; // buffer pointer 4 longs

HUBDATA int GbI2C_result ;

/*
=================================================
start i2c devices
This method is used to start the driver.
Parameter semptr is a lock semaphore to be used by the driver to lock SCL
scl_pin and sda_pin are the pin number to use for the I2C interface.
=================================================
*/

int GbI2C_start(void)
{
    extern unsigned int I2cDevices_array[];
    GbI2C_command   = GbI2C_scl_pin | (GbI2C_sda_pin << 16);
    GbI2C_cog       = cognew(I2cDevices_array, &GbI2C_command) + 1;

    while (GbI2C_command)
        ; // wait for cog start

    GbI2C_rtc_writeReg(GbI2C_RTC_CONTROL, 0x80);    // set LCD on with 1338 OUT

    GbI2C_accel_writeReg(GbI2C_ACCEL_MODE, 0);      // clear mode bit for register writes
    GbI2C_accel_writeReg(GbI2C_ACCEL_SLPCNT, 0);    // no sleep count
    GbI2C_accel_writeReg(GbI2C_ACCEL_PDET, 0xe0);   // no tap detect
    GbI2C_accel_writeReg(GbI2C_ACCEL_SRFLT, 0x34);  // sample rate 8 sps / filter = 2
    GbI2C_accel_writeReg(GbI2C_ACCEL_MODE, 1);      // enable active mode

    return GbI2C_cog;
}

/*
=================================================================
This method is used to stop the driver.
=================================================================
*/
void GbI2C_stop(void)
{
    if (GbI2C_cog > 0) {
        cogstop(GbI2C_cog-1);
        GbI2C_cog = 0;
    }
}

/*
=================================================================
This method reads the device register via the PASM code.
=================================================================
*/
int GbI2C_accel_readReg(int maddr)
{
    GbI2C_device  = GbI2C_ACCEL_ID;
    GbI2C_address = 0;
    GbI2C_bufptr  = (int*)&GbI2C_accelbuf;
    GbI2C_command = GbI2C_CMD_ACCEL_READ;
    while (GbI2C_command);
    return GbI2C_accelbuf[maddr];
}

/*
=================================================================
This method writes the device register via the PASM code.
=================================================================
*/
void GbI2C_accel_writeReg(int maddr, unsigned char mdata)
{
    GbI2C_result  = mdata;
    GbI2C_device  = GbI2C_ACCEL_ID;
    GbI2C_address = maddr;
    GbI2C_bufptr  = &GbI2C_result;
    GbI2C_command = GbI2C_CMD_ACCEL_WRITE;
    while (GbI2C_command);
}

/*
=================================================================
Method reads the ADC SMOOTHER times and returns average voltage * 100
=================================================================
*/
int GbI2C_adc_getVoltage100(void)
{
    int n;
    int result = 0;
    for (n = 0; n < GbI2C_SMOOTHER; n++) {
        result += GbI2C_adc_read();
    }
    result = result/(GbI2C_SMOOTHER+1);
    return result;
}

/*
=================================================================
This method reads the device register via the PASM code.
=================================================================
*/
int GbI2C_adc_read(void)
{
    GbI2C_device  = GbI2C_ADC_ID;
    GbI2C_bufptr  = (int*)&GbI2C_result;  // need to convert to hubaddr
    GbI2C_command = GbI2C_CMD_ADC_READ;
    while (GbI2C_command);
    return GbI2C_result;
}

/*
=================================================================
Use this method to read a buffer from EEPROM
=================================================================
*/
void GbI2C_ee_readBuffer(int maddr, unsigned char *buffptr, int mlen)
{
    GbI2C_device  = GbI2C_EEPROM_ID;
    GbI2C_address = maddr;
    GbI2C_count   = mlen;
    GbI2C_bufptr  = (int*)buffptr;
    GbI2C_command = GbI2C_CMD_EEPROM_RDBUF;
    while (GbI2C_command);
}

/*
=================================================================
Use this method to write a buffer to EEPROM
=================================================================
*/
void GbI2C_ee_writeBuffer(int maddr, unsigned char *buffptr, int mlen)
{
    GbI2C_device  = GbI2C_EEPROM_ID;
    GbI2C_address = maddr;
    GbI2C_count   = mlen;
    GbI2C_bufptr  = (int*)buffptr;
    GbI2C_command = GbI2C_CMD_EEPROM_WRBUF;
    while (GbI2C_command);
}


/*
=================================================================
Use this method to read one byte from EEPROM
=================================================================
*/
unsigned char GbI2C_ee_readByte(int maddr)
{
    GbI2C_ee_readBuffer(maddr, GbI2C_eeBuffer, 1);
    return GbI2C_eeBuffer[0];
}


/*
=================================================================
Use this method to write one byte to EEPROM
=================================================================
*/
void GbI2C_ee_writeByte(int maddr, unsigned char value)
{
    GbI2C_eeBuffer[0] = value;
    GbI2C_ee_writeBuffer(maddr, GbI2C_eeBuffer, 1);
}


/*
=================================================================
This method gets a device register saved to the rtcbuf
=================================================================
*/
int GbI2C_rtc_getReg(int maddr)
{
    return GbI2C_rtcbuf[maddr];
}

/*
=================================================================
This method reads the device register via the PASM code.
=================================================================
*/
void GbI2C_rtc_readRegs(void)
{
    GbI2C_device  = GbI2C_RTC_ID;
    GbI2C_address = 0;
    GbI2C_bufptr  = (int*)&GbI2C_rtcbuf;
    GbI2C_command = GbI2C_CMD_RTC_READ;
    while (GbI2C_command);
}

/*
=================================================================
This method reads the device register via the PASM code.
=================================================================
*/
int GbI2C_rtc_readReg(int maddr)
{
    GbI2C_device  = GbI2C_RTC_ID;
    GbI2C_address = 0;
    GbI2C_bufptr  = (int*)&GbI2C_rtcbuf;
    GbI2C_command = GbI2C_CMD_RTC_READ;
    while (GbI2C_command);
    return GbI2C_rtcbuf[maddr];
}

/*
=================================================================
This method writes the device register via the PASM code.
=================================================================
*/
void GbI2C_rtc_writeReg(int maddr, unsigned char mdata)
{
    GbI2C_result  = mdata;
    GbI2C_device  = GbI2C_RTC_ID;
    GbI2C_address = maddr;
    GbI2C_bufptr  = &GbI2C_result;
    GbI2C_command = GbI2C_CMD_RTC_WRITE;
    while (GbI2C_command);
}


#if 0
/*
=================================================
 return status of all buttons
*/
int GbButtons_buttons(void)
{
    return GbI2C_buttons;
}

/*
=================================================
 this returns state of a/b buttons as 2 bits

 bit 0 is state of a : 1 pressed, 0 released
 bit 1 is state of b : 1 pressed, 0 released
*/
int GbButtons_abstate(void)
{
    return (GbI2C_buttons >> 1) & 3;
}

/*
=================================================
 this returns state of direction button as a bit mask

 0x0 : all      released
 0x1 : up       pressed
 0x2 : down     pressed
 0x4 : left     pressed
 0x8 : right    pressed
*/
int GbButtons_direction(void)
{
    return (GbI2C_buttons >> 5) & 0xf;
}

/*
=================================================
 this returns state of the select button
 1 start pressed, 0 start released
*/
int GbButtons_selectState(void)
{
    return (GbI2C_buttons >> 3) & 1;
}

/*
=================================================
 this returns state of the start button
 1 start pressed, 0 start released
*/
int GbButtons_startState(void)
{
    return (GbI2C_buttons >> 4) & 1;
}

/*
=================================================
 this returns state of the enter button
 1 start pressed, 0 start released
*/
int GbButtons_enterState(void)
{
    return GbI2C_buttons & 1;
}

#endif
