/*
==================================================
GUI.c
This module provides GUI elements.

Elements include:
    Panel, Label, Button, 3Dbutton, TextBox (tbd), ListBox (tbd) 

Elements are geometric shapes drawn with specific
colors. Text printed to a shape after drawing will
go to a default position. Some GUI elements like
ListBox have more than one position.

==================================================
*/
#include <string.h>
#include "LCD.h"
#include "GUI.h"

int GUI_xpos;
int GUI_ypos;
int GUI_TextOffset    = 8;

int GUI_FaceColor     = LCD_WHITE_GRAY;
int GUI_GRADIENT      = 0x842;

/*
 * ==================================================
 * draw a simple button
 * @param text - caption for button
 * @param x - left start position
 * @param y - top start position
 */ 
void GUI_button(char* text, int x, int y)
{
    int n, w, h, mlen;
    n = strlen(text);
    mlen = 3;
    if(n < mlen) n = mlen;
    w = LCD_FONTOFFSET+n*(LCD_FONTWIDTH+2);
    h = LCD_FONTOFFSET+1+LCD_FONTHEIGHT*2;
    GUI_button_parms(text,LCD_BLACK,GUI_FaceColor,x,y,w,h,0);
}

/* ==================================================
 *  draw a simple 3D button
 *  @param text - caption for button
 *  @param x - left start position
 *  @param y - top start position
 */
void GUI_3Dbutton(char *text, int x, int y)
{
    int n, w, h, mlen;
    n = strlen(text);
    mlen = 3;
    if(n < mlen) n = mlen;
    w = LCD_FONTOFFSET+n*(LCD_FONTWIDTH+2);
    h = LCD_FONTOFFSET+1+LCD_FONTHEIGHT*2;
    GUI_button_parms(text,LCD_BLACK,GUI_FaceColor,x,y,w,h,1);
}


/* ==================================================
 *  draw a button of color, width and height to x,y.
 *  @param thrd - set for 3D effect
 *  @param text - button caption
 *  @param fgcolor - text color
 *  @param bgcolor - button background or face color
 *  @param x - left side
 *  @param y - right side
 *  @param w - width
 *  @param h - height
 */
void GUI_button_parms(char * text, int fgcolor, int bgcolor, int x, int y, int w, int h, int thrd)
{
    int n = 0;
    int dy = 0;
    int adj = 0;

    LCD_setFgColor(fgcolor);
    LCD_setBgColor(bgcolor);

    if(thrd) {
        adj = 0;
        dy = 0;
        LCD_fillRectangle(LCD_WHITEST_GRAY,x,y,w,h);
        for(n = 0; n <  h/2-5; n++) {
            LCD_fillRectangle(LCD_LIGHT_GRAY+adj,x,y+dy,w,1);
            adj = adj + GUI_GRADIENT;
            dy++;
        }
        dy = dy + 10;
        for(n = 0; n < h/2-5; n++) {
            adj = adj - GUI_GRADIENT;
            LCD_fillRectangle(LCD_LIGHT_GRAY+adj,x,y+dy,w,1);
            dy++;
        }
    }
    else {
        LCD_fillRectangle(bgcolor,x,y,w,h);
    }

    LCD_fillRectangle(LCD_WHITE,x,y,1,h);                 // left line
    LCD_fillRectangle(LCD_DARK_GRAY,x+w,y,1,h);           // right line
    LCD_fillRectangle(LCD_DARK_GRAY,x+1,y+h,w,1);         // bottom line
    LCD_fillRectangle(LCD_WHITE,x,y,w,1);                 // top line
/*
    drawRectangle is broken ????
    LCD_drawRectangle(LCD_WHITE,x,y,1,h,1)                // left line
    LCD_drawRectangle(LCD_DARK_GRAY,x+w,y,1,h,1)          // right line
    LCD_drawRectangle(LCD_DARK_GRAY,x+1,y+h,w,2,1)        // bottom line
    LCD_drawRectangle(LCD_WHITE,x,y,w,1,1)                // top line
*/
    GUI_xpos = x + GUI_TextOffset;
    GUI_ypos = y + GUI_TextOffset;

    LCD_setBgColor(LCD_TRANSPARENT);
    LCD_drawString(text, GUI_xpos, GUI_ypos);
}

/* ==================================================
 *  draw a standard color panel, width and height at x,y.
 */
void GUI_panel(int x, int y, int w, int h)
{
    GUI_panel_parms(GUI_FaceColor,x,y,w,h,0);
}

/* ==================================================
 *  draw a standard color panel, width and height at x,y.
 */
void GUI_panelEtched(int x, int y, int w, int h)
{
    GUI_panel_parms(GUI_FaceColor,x,y,w,h,8);
}

/* ==================================================
 *  draw a panel of bgcolor, width and height at x,y.
 */
void GUI_panel_parms(int bgcolor, int x, int y, int w, int h, int etch)
{
    LCD_setBgColor(bgcolor);
    LCD_fillRectangle(bgcolor,x,y,w,h);                 // rectangle
    LCD_fillRectangle(LCD_WHITE,x,y,w,2);               // top line
    LCD_fillRectangle(LCD_WHITE,x,y,1,h+2);             // left line
    LCD_fillRectangle(LCD_DARK_GRAY,x+w-1,y+1,1,h-1);   // right line
    LCD_fillRectangle(LCD_DARK_GRAY,x+1,y+h,w-1,3);     // bottom line
    if(etch != 0) {
        if(w>2*etch) {
            if(h>2*etch+2)
                LCD_drawRectangle(LCD_GRAY,x+etch,y+etch,w-2*etch,h-2*etch,1);
        }
    }
    GUI_xpos = x + GUI_TextOffset + etch;
    GUI_ypos = y + GUI_TextOffset - etch/2 + 1;
}

/* ==================================================
 *  print panel label
 *  user is responsible for length
 */
void GUI_panelLabel(char *text)
{
    LCD_drawString(text, GUI_xpos, GUI_ypos);
}

/* ==================================================
 *  print text label
 *  user is responsible for length
 */
void GUI_label(char *text, int x, int y)
{
    LCD_drawString(text, x, y);
}

/* ==================================================
 *  print underlined text label
 *  user is responsible for length
 */
void GUI_labelUL(char *text, int x, int y)
{
    int n;
    LCD_drawString(text, x, y);
    n = strlen(text);
    LCD_fillRectangle(LCD_fgcolor,x,y+10,n*LCD_FONTWIDTH,1);  // top line
}

/* ==================================================
 *  print text to gui residual xpos, ypos
 *  user is responsible for length
 */
void GUI_text(char *text)
{
    LCD_drawString(text, GUI_xpos, GUI_ypos);
}

/* ==================================================
 *  set foreground text color
 */
void GUI_textColor(int color)
{
    LCD_setFgColor(color);
}

