{{
=====================================================================
I2cDevices.spin - GameBaby I2C devices driver.
Attempts to combign all GameBaby I2C devices into one file.
Devices include: EEPROM, accelerometer, battery ADC, and RTC.
=====================================================================
}}

con
{{
=====================================================================
PASM command interface
=====================================================================
}}
    CMD_ADC_READ        = 1
    CMD_ACCEL_READ      = 2
    CMD_ACCEL_WRITE     = 3
    CMD_EEPROM_RDBUF    = 4
    CMD_EEPROM_WRBUF    = 5
    CMD_RTC_READ        = 6
    CMD_RTC_WRITE       = 7

    CMDMASK             = $7    ' one of 7 states

con
    scl_pin     =       28      ' always 28. Shared with Buttons/Flash CLK
    sda_pin     =       29      ' always 29. Not shared.

    ACCEL_ID    =       $98     ' Accelerometer device ID
    ADC_ID      =       $9A     ' ADC device ID
    EEPROM_ID   =       $A0     ' EEPROM device ID
    RTC_ID      =       $D0     ' RTC device ID

    EE_PAGELEN  =       128     ' AT24C512 EEPROM PAGE LENGTH
    RBIT        =       1       ' devid byte read bit

con ' delays to accomodate various I2C speeds
    ADC_DELAY       = 1
    ACCEL_DELAY     = 1
    EEPROM_DELAY    = 1
    RTC_DELAY       = 50 

con
{{
=====================================================================
ACCELEROMETER REGISTERS
=====================================================================
Address Name Definition         Bit 7   Bit 6   Bit 5   Bit 4   Bit 3   Bit 2   Bit 1   Bit 0
    $00 XOUT 6-bit output X     -       Alert   XOUT[5] XOUT[4] XOUT[3] XOUT[2] XOUT[1] XOUT[0]
    $01 YOUT 6-bit output Y     -       Alert   YOUT[5] YOUT[4] YOUT[3] YOUT[2] YOUT[1] YOUT[0]
    $02 ZOUT 6-bit output Z     -       Alert   ZOUT[5] ZOUT[4] ZOUT[3] ZOUT[2] ZOUT[1] ZOUT[0]
    $03 TILT Tilt Status        Shake   Alert   Tap     PoLa[2] PoLa[1] PoLa[0] BaFro[1] BaFro[0]
    $04 SRST Sampling Rate Stat 0       0       0       0       0       0       AWSRS   AMSRS
    $05 SPCNT Sleep Count       SC[7]   SC[6]   SC[5]   SC[4]   SC[3]   SC[2]   SC[1]   SC[0]
    $06 INTSU Interrupt Setup   SHINTX  SHINTY  SHINTZ  GINT    ASINT   PDINT   PLINT   FBINT
    $07 MODE Mode               IAH     IPP     SCPS    ASE     AWE     TON     -       MODE
    $08 SR Filter               FILT[2] FILT[1] FILT[0] AWSR[1] AWSR[0] AMSR[2] AMSR[1] AMSR[0]
    $09 PDET Tap Detection      ZDA     YDA     XDA     PDTH[4] PDTH[3] PDTH[2] PDTH[1] PDTH[0]
    $0A PD Tap Debounce Count   PD[7]   PD[6]   PD[5]   PD[4]   PD[3]   PD[2]   PD[1]   PD[0]
=====================================================================
}}
    ACCEL_XOUT    = 0
    ACCEL_YOUT    = 1
    ACCEL_ZOUT    = 2
    ACCEL_TILT    = 3
    ACCEL_SRST    = 4
    ACCEL_SLPCNT  = 5
    ACCEL_INTSU   = 6
    ACCEL_MODE    = 7
    ACCEL_SRFLT   = 8
    ACCEL_PDET    = 9
    ACCEL_PDCNT   = 10

{{
=====================================================================
RTC REGISTERS
=====================================================================
}}
    RTC_SECONDS     = 0
    RTC_MINUTES     = 1
    RTC_HOURS       = 2
    RTC_DAY         = 3
    RTC_DATE        = 4
    RTC_MONTH       = 5
    RTC_YEAR        = 6
    RTC_CONTROL     = 7

{{
=====================================================================
PASM Command interface
=====================================================================
}}
dat
    cog         long    0       ' our cog number

    command     long    0       ' command startup with scl & sda pins
    semaddr     long    0       ' semaphore pointer
    device      long    0       ' default device number
    address     long    0       ' register address
    count       long    0       ' operation data length
    buffer      long    0 [4]   ' buffer pointer

'#define BMAD

pub start(semptr)
{{
=====================================================================
This method is used to start the driver.
Parameter semptr is a lock semaphore to be used by the driver to lock SCL
scl_pin and sda_pin are the pin number to use for the I2C interface.
=====================================================================
}}
    semaddr := semptr
    command := scl_pin | sda_pin << 16 

#ifndef BMAD
    cog := cognew(@entry, @command) + 1
    repeat until command == 0

    accel_writeReg(ACCEL_MODE, 0)          ' clear mode bit for register writes
    accel_writeReg(ACCEL_SLPCNT, 0)        ' no sleep count
    accel_writeReg(ACCEL_PDET, $e0)        ' no tap detect
    accel_writeReg(ACCEL_SRFLT, $34)       ' sample rate 8 sps / filter = 2
    accel_writeReg(ACCEL_MODE, 1)          ' enable active mode

#else
    cognew(runit,@stack)
    d.debug(@entry,@command)

obj d : "BMAUtility"
var long stack[$40]
var byte testbuffer[$400]

pub runit
    repeat
'       adc_getVoltage100
'       accel_readReg(0)
'       ee_readByte(0)
'       ee_readBuffer(0,@testbuffer, 130)
        rtc_readRegs

#endif

pub stop
{{
=====================================================================
This method is used to stop the driver.
=====================================================================
}}
    if cog > 0
        cogstop(cog~ -1)

dat
    accelbuf    long 0 [4]

pub accel_readReg(maddr)
{{
=====================================================================
This method reads the device register via the PASM code.
=====================================================================
}}
    device  := ACCEL_ID   ' initialized in spin struct
    address := 0
    buffer  := @accelbuf
    command := CMD_ACCEL_READ
    repeat until command == 0
    return byte[@accelbuf][maddr]

pub accel_writeReg(maddr,mdata)
{{
=====================================================================
This method writes the device register via the PASM code.
=====================================================================
}}
    device  := ACCEL_ID   ' initialized in spin struct
    address := maddr
    buffer  := @mdata
    command := CMD_ACCEL_WRITE
    repeat until command == 0

pub adc_getVoltage100 | n
{{
=====================================================================
Method reads the ADC SMOOTHER times and returns average voltage * 100
=====================================================================
}}
    result := 0
    repeat n from 0 to SMOOTHER-1
        result += adc_read
    result /= SMOOTHER
con
    SMOOTHER    = 40


pub adc_read
{{
=====================================================================
This method reads the device register via the PASM code.
=====================================================================
}}
    device  := ADC_ID
    buffer  := @result
    command := CMD_ADC_READ
    repeat until command == 0


pub ee_readBuffer(maddr, buffptr, mlen)
{{
=====================================================================
Use this method to read a buffer from EEPROM
=====================================================================
}}
    device  := EEPROM_ID
    address := maddr
    count   := mlen
    buffer  := buffptr
    command := CMD_EEPROM_RDBUF
    repeat until command == 0

pub ee_writeBuffer(maddr, buffptr, mlen)
{{
=====================================================================
Use this method to write a buffer to EEPROM
=====================================================================
}}
    device  := EEPROM_ID
    address := maddr
    count   := mlen
    buffer  := buffptr
    command := CMD_EEPROM_WRBUF
    repeat until command == 0

con
    eeBufferLen = $20
'dat eeBuffer  byte 0 [eeBufferLen]
var byte eeBuffer[eeBufferLen]

pub ee_readByte(maddr)
{{
=====================================================================
Use this method to read one byte from EEPROM
=====================================================================
}}
    ee_readBuffer(maddr, @eeBuffer, 1)
    return eeBuffer[0]

pub ee_writeByte(maddr,value)
{{
=====================================================================
Use this method to write one byte to EEPROM
=====================================================================
}}
    eeBuffer[0] := value
    ee_writeBuffer(maddr, @eeBuffer, 1)

dat
    rtcbuf    long 0 [$10]

pub rtc_getReg(maddr)
    return byte[@rtcbuf][maddr]

pub rtc_readRegs
{{
=====================================================================
This method reads the device register via the PASM code.
=====================================================================
}}
    device  := RTC_ID   ' initialized in spin struct
    address := 0
    buffer  := @rtcbuf
    command := CMD_RTC_READ
    repeat until command == 0

pub rtc_readReg(maddr)
{{
=====================================================================
This method reads the device register via the PASM code.
=====================================================================
}}
    device  := RTC_ID   ' initialized in spin struct
    address := 0
    buffer  := @rtcbuf
    command := CMD_RTC_READ
    repeat until command == 0
    return byte[@rtcbuf][maddr]

pub rtc_writeReg(maddr,mdata)
{{
=====================================================================
This method writes the device register via the PASM code.
=====================================================================
}}
    device  := RTC_ID   ' initialized in spin struct
    address := maddr
    buffer  := @mdata
    command := CMD_RTC_WRITE
    repeat until command == 0

pub cmdptr
{{
=====================================================================
Return pointer to command struct for initializing with another PASM driver.
=====================================================================
}}
    return @command
  
pub pasm
{{
=====================================================================
Return pointer to PASM struct for initializing with another PASM driver.
=====================================================================
}}
    return @entry

pub size
{{
=====================================================================
Return size of PASM.
=====================================================================
}}
    return @eof-@entry

{{
=====================================================================
PASM Driver
=====================================================================
}}
dat             org 0
entry
long 0 [8]

_command        mov     $, par          ' command register
_tmp            mov     $, par          ' command register
_scl            add     _tmp, #4
_semaddr        mov     $, _tmp         ' semaphore pointer
_sda            add     _tmp, #4
_device         mov     $, _tmp         ' default device number
_ack            add     _tmp, #4
_address        mov     $, _tmp         ' register address
_val            add     _tmp, #4
_length         mov     $, _tmp         ' operation data length
_cmd            add     _tmp, #4
_buffer         mov     $, _tmp         ' buffer pointer
_lockID         rdlong  $, _semaddr     ' read lock ID for sharing SCL
_t0             mov     _scl, #1
_t1             shl     _scl, #SCL_PIN
_addr           rdword  _t0, _tmp
ndx             mov     _sda, #1
_pagesize       shl     _sda, #SDA_PIN
_dev            rdlong  _lockId, _lockId
_bptr           jmp     #_cmdZero       ' must jump over lock clear

'---------------------------------------------------------------
' release semaphore and signal command ready
'
_cmdDone        lockclr _lockID         ' release lock on cmdDone
_cmdZero        wrlong  _zero, _command

'---------------------------------------------------------------
' wait for command
'
_cmdLoop        rdlong  _cmd, _command wz
'   if_z        jmp     #_cmdLoop
    if_nz       jmp     #:gotcommand        ' skip sleep if_nz
                mov     _t1, _bedtime
                add     _t1, cnt
                waitcnt _t1, _t1               ' COG sleep saves energy
                jmp     #_cmdLoop

:gotcommand
'---------------------------------------------------------------
' wait for SCL to be unlocked here
'
:getlock        lockset _lockID wz
    if_nz       jmp     #:getlock

'---------------------------------------------------------------
' calculate jump to handler
'
                rdlong  _bytes, _length
                movs    :jump, #_jumptable   ' get jumpbase
                rdlong  _bptr, _buffer      ' data is passed by reference
                and     _cmd, #CMDMASK       ' one of 7 states
                shl     _cmd, #1             ' commands are 2 longs each
                rdlong  _addr, _address     ' address
                add     :jump, _cmd
                rdlong  _dev, _device       ' data is passed by reference
:jump           jmp     #0-0


'               call    #_scopeDelay
' delay used just for scope loop
{
_scopeDelay
                mov     _t1, #$100
                shl     _t1, #8
                add     _t1, cnt
                waitcnt _t1, _t1
_scopeDelay_ret ret
'}

'---------------------------------------------------------------
' command jump table    - table could be single instruction
'                       jump to handler to reduce overhead.
'
_jumptable      nop
                jmp     #_cmdDone

_cmdAdcREAD             ' = 1
                call    #_AdcREAD
                jmp     #_cmdDone


'---------------------------------------------------------------
'
_cmdAccelREAD           ' = 2
                call    #_AccelREAD
                jmp     #_cmdDone

'---------------------------------------------------------------
'
_cmdAccelWRITE          ' = 3
                call    #_AccelWRITE
                jmp     #_cmdDone

'---------------------------------------------------------------
'
_cmdEepromRDBUF         ' = 4
                call    #_EepromRDBUF
                jmp     #_cmdDone

'---------------------------------------------------------------
'
_cmdEepromWRDBUF        ' = 5
                call    #_EepromWRBUF
                jmp     #_cmdDone

'---------------------------------------------------------------
'
_cmdRtcREAD             ' = 6
                call    #_RtcREAD
                jmp     #_cmdDone

'---------------------------------------------------------------
'
_cmdRtcWRITE            ' = 7
                call    #_RtcWRITE
                jmp     #_cmdDone

'---------------------------------------------------------------
' @param       _dev  - device number
' @param       _addr - read address
' @param       _bptr - address of buffer for read
'
_AdcREAD
                mov     _delayCount,#ADC_DELAY
'
' send write device command
'
                call    #_sendStart          ' start then put address in byte 0 for _sendVal
                or      _dev, #RBIT          ' send write command
                mov     _val, _dev
                call    #_sendVal            ' send device id and upper address bits
{
' optional delay for ADC to finish. seems to not matter for this device
                mov     _t1, #$100
                shl     _t1, #4
                add     _t1, cnt
                waitcnt _t1, _t1               ' wait for ADC to finish before read
}

'
' receive data
'
                call    #_recVal             ' read data    - just use 8 bit mode if possible
                mov     _addr, _val
                shl     _addr, #8
                call    #_recVal             ' read data    - just use 8 bit mode if possible
                call    #_sendEnd
                or      _addr, _val
                mov     _val, _addr
                shr     _val, #2              ' valid data is in bit 2^2 through 2^12
'
' write data back to user
'
                wrword  _val, _bptr
_AdcREAD_ret    ret


'---------------------------------------------------------------
' @param       _dev  - device number
' @param       _addr - read address
' @param       _bptr - address of buffer for read
'
_AccelREAD

                mov     _delayCount,#ACCEL_DELAY
'
' send write device command
'
                call    #_sendWrStart
'
' send start again to specify a read command
'
                call    #_sendStart  ' start then put address in byte 0 for _sendVal
                or      _dev, #RBIT  ' send read command
                mov     _val, _dev
                call    #_sendVal    ' send device id and upper address bits

' receive data
'
                andn    _ack, _sda   ' set ack bit so the device knows were done for single byte read
'{
                mov     _t1, #10
:loop
                call    #_recVal     ' read data    - just use 8 bit mode if possible
                wrbyte  _val, _bptr
                add     _bptr, #1
                djnz    _t1, #:loop
'}
                or      _ack, _sda   ' set ack bit so the device knows were done for single byte read
                call    #_recVal     ' read data    - just use 8 bit mode if possible
                call    #_sendEnd
'
' write data back to user
'
                wrbyte  _val, _bptr
_AccelREAD_ret  ret

'---------------------------------------------------------------
' write byte to address
' @param       _dev - device number
' @param       addr - read address
' @param       _bptr   - address of buffer for write
'
_AccelWRITE
                mov     _delayCount,#ACCEL_DELAY
'
' send write device command
'
                call    #_sendWrStart

' write data

                or      _ack, _sda           ' set ack bit so the device knows were done for single byte write
                rdbyte  _val, _bptr         ' get byte to write
                call    #_sendVal            ' read data    - just use 8 bit mode if possible
                call    #_sendEnd
_AccelWRITE_ret ret


'---------------------------------------------------------------
' @param       _dev  - device number
' @param       _addr - read address
' @param       _bptr - address of buffer for read
'
_RtcREAD
                mov     _delayCount,#RTC_DELAY
'
' send start to write address
'
                call    #_sendWrStart
'
' send repeated start to specify a read command
'
                call    #_sendStart  ' start then put address in byte 0 for _sendVal
                or      _dev, #RBIT  ' send read command
                mov     _val, _dev
                call    #_sendVal    ' send device id and upper address bits

' receive data
'
                andn    _ack, _sda   ' clear ack bit so the device knows we need multiple bytes
'{
                mov     _t1, #$3f
:loop
                call    #_recVal     ' read data    - just use 8 bit mode if possible
                wrbyte  _val, _bptr
                add     _bptr, #1
                djnz    _t1, #:loop
'}
                or      _ack, _sda   ' set ack bit so the device knows were done for single byte read
                call    #_recVal     ' read data    - just use 8 bit mode if possible
                call    #_sendEnd
'
' write data back to user
'
                wrbyte  _val, _bptr
_RtcREAD_ret    ret

'---------------------------------------------------------------
' write byte to address
' @param       _dev - device number
' @param       addr - read address
' @param       _bptr   - address of buffer for write
'
_RtcWRITE
                mov     _delayCount,#RTC_DELAY
'
' send write device command
'
                call    #_sendWrStart

' write data

                or      _ack, _sda           ' set ack bit so the device knows were done for single byte write
                rdbyte  _val, _bptr         ' get byte to write
                call    #_sendVal            ' read data    - just use 8 bit mode if possible
                call    #_sendEnd
_RtcWRITE_ret   ret

'
' send write start device command routine
'
_sendWrStart
                call    #_sendStart          ' start then put address in byte 0 for _sendVal

                andn    _dev, #RBIT          ' send write command
                mov     _val, _dev
                call    #_sendVal            ' send device id and upper address bits

                mov     _val, _addr
                call    #_sendVal            ' send register address
_sendWrStart_ret ret


{{
'-------------------------------------------------------------------------------
' read buffer from eeprom
' @param       _dev - device number
' @param       addr - read address
' @param       _bytes - number of bytes to read
' @param       _bptr   - address of buffer for read
'
}}
_EepromRDBUF
                mov     _delayCount,#EEPROM_DELAY
:loop
                cmp     _bytes, #EE_PAGELEN wc,wz
  if_be         mov     _pagesize, _bytes
  if_a          mov     _pagesize, #EE_PAGELEN
                call    #_readPage
                add     _addr, _pagesize
                sub     _bytes, _pagesize wc,wz
  if_a          jmp     #:loop
_EepromRDBUF_ret ret

{{
'-------------------------------------------------------------------------------
' write buffer to eeprom
' @param       *dp - device number
' @param       *ap - write address
' @param       *sp - number of bytes to write
' @param       *bp - address of buffer to write
'
}}
_EepromWRBUF
                mov     _delayCount,#EEPROM_DELAY
:loop
                cmp     _bytes, #EE_PAGELEN wc,wz
    if_be       mov     _pagesize, _bytes
    if_a        mov     _pagesize, #EE_PAGELEN
                call    #_writePage
                add     _addr, _pagesize
                sub     _bytes, _pagesize wc,wz
    if_a        jmp     #:loop
_EepromWRBUF_ret ret


'-------------------------------------------------------------------------------
' read eeprom page
' @param       *dp - device number
' @param       *ap - read address
' @param       *bp - address of buffer for read
' @param       _pagesize - number of bytes to read
'
_readPage
                call         #_sendAddr
                or           _dev,#RBIT         ' send read command
                call         #_sendCurrent
                mov          _t1, _pagesize          ' read _pagesize bytes
                sub          _t1, #1            ' first _pagesize-1 are NAK
                tjz          _t1, #:last
                andn         _ack, _sda          ' not last byte     
:loop
                call         #_recVal            ' read page data
                wrbyte       _val, _bptr
                add          _bptr, #1
                djnz         _t1, #:loop
:last  
                or           _ack, _sda         ' last byte has ACK     
                call         #_recVal
                wrbyte       _val, _bptr
                add          _bptr, #1          ' increment for next loop too
                call         #_sendEnd
_readPage_ret   ret

{{
'-------------------------------------------------------------------------------
' read buffer from eeprom with early ack after 1 byte read
' @param       _dev - device number
' @param       _addr - read address
' @param       bytes - number of bytes to read
' @param       _bptr   - address of buffer for read
'
}}
{
cmdReadBytes
                'mov        _pagesize,    _bytes
                'call       #_readPage
                call        #_sendAddr
                or          _dev,    #RBIT   ' send read command
                call        #_sendCurrent
                mov         _t1,      _bytes   ' read _pagesize bytes
                'sub        _t1,      #1      ' first _pagesize-1 are NAK
                tjz         _t1,      #:last
                andn        _ack,     _sda     ' not last byte
:loop
                call        #_recVal          ' read page data
                wrbyte      _val,     _bptr
                add         _bptr,      #1
                djnz        _t1,      #:loop
:last
                or          _ack,     _sda     ' last byte has ACK
                call        #_recVal
                wrbyte      val,     _bptr
                add         _bptr,      #1      ' increment for next loop too
                call        #_sendEnd
                jmp         #cmdDone
}

'-------------------------------------------------------------------------------
' read current byte
' @param       *dp - device number
' @param       *ap - read address
' @param       *sp - number of bytes to read
' @param       *bp - address of buffer for read
'
_readCurrent
                or           _dev, #RBIT       ' read
                call         #_sendCurrent
                or           _ack,  _sda        ' last byte     
                call         #_recVal
                wrbyte       _val,  _bptr
                add          _bptr, #1
                call         #_sendEnd
_readCurrent_ret   ret

'-------------------------------------------------------------------------------
' write eeprom page
' @param       *dp - device number
' @param       *ap - write address
' @param       *bp - address of buffer for write
' @param       _pagesize - number of bytes to read
'
_writePage
                call         #_sendAddr          ' send page address
                mov          _t1,  _pagesize
:loop
                rdbyte       _val, _bptr
                add          _bptr,#1
                call         #_sendVal           ' send data
                djnz         _t1,  #:loop
                call         #_writeEndPoll
_writePage_ret  ret

'-------------------------------------------------------------------------------
' send end and poll for write complete
'
_writeEndPoll
                call         #_sendEnd           ' send stop
                call         #_sclDelay
                call         #_sendCurrent       ' now start write polling
                or           outa, _scl
                andn         dira, _sda         ' looking for ACK
:loop
                test         _sda, ina wz       ' is ack 0 ?
                if_nz jmp    #:loop             ' if ack 1, repeat
_writeEndPoll_ret  ret  

'-------------------------------------------------------------------------------
' write current eeprom byte
' @param       *dp - device number
' @param       *ap - write address
' @param       *sp - number of bytes to write
' @param       *bp - address of buffer for write
'
_writeCurrent
                andn         _dev, #RBIT
                call         #_sendStart
                rdbyte       _val, _bptr
                add          _bptr,#1
                call         #_sendVal
_writeCurrent_ret  ret

'-------------------------------------------------------------------------------
' send eeprom address
' @param       _dev - device number
' @param       _addr - read address

'
_sendAddr
                call         #_sendStart
                mov          _val,    _addr     ' %00000000_00000111_11111111_11111111 ->
                ror          _val,    #16       ' %11111111_11111111_00000000_00000111
                shl          _val,    #1        ' %11111111_11111110_00000000_00001110 ->
                andn         _dev,    #RBIT     ' %10100000 send addr/write needs bit 0 low 
                or           _val,    _dev      ' %11111111_11111110_00000000_10101110
                call         #_sendVal           ' send device id and upper address bits
                mov          _val,    _addr     ' %00000000_00000111_11111111_11111111 ->
                ror          _val,    #8        ' %11111111_00000000_00000111_11111111
                call         #_sendVal           ' send middle address bits
                mov          _val,    _addr     ' %00000000_00000111_11111111_11111111
                call         #_sendVal           ' send lower address bits
_sendAddr_ret   ret

'-------------------------------------------------------------------------------
' send current is used to initiate a transaction after _sendAddr
' @param       devs - device number with bit 0 set for read or clear for write
' @param       addr - read address
'
_sendCurrent
                call         #_sendStart
                mov          _val,    _addr     ' %00000000_00000111_11111111_11111111 ->
                ror          _val,    #16       ' %11111111_11111111_00000000_00000111
                shl          _val,    #1        ' %11111111_11111110_00000000_00001110 ->
                or           _val,    _dev      ' %11111111_11111110_00000000_1010111r
                call         #_sendVal           ' send device id and upper address bits
_sendCurrent_ret  ret


'---------------------------------------------------------------
' A start signal is when SDA goes 1 -> 0 while SCL is 1
'
_sendStart
                or      outa,    _scl            ' scl high
                or      dira,    _scl            ' scl output ... but it is always output right?
                or      outa,    _sda            ' sda high
                or      dira,    _sda            ' sda output
                call    #_sclDelayD2              ' device detects high
                andn    outa,    _sda            ' sda low for start condition 
                call    #_sclDelayD2              ' device detects start
                andn    outa,    _scl            ' finish with scl low 
                call    #_sclDelayD2              ' device detects start
_sendStart_ret   ret
  
'---------------------------------------------------------------
' An end signal is when SDA goes 0 -> 1 while SCL is 1
'
_sendEnd
                andn    outa,    _sda            ' sda high
                or      outa,    _scl            ' scl high
                call    #_sclDelayD2
                or      outa,    _sda            ' sda high
                call    #_sclDelayD2              ' device detects stop
                andn    dira,    _sda            
                andn    dira,    _scl            ' leave sda & scl input so other devices can use the bus
_sendEnd_ret     ret

'---------------------------------------------------------------
' scl has limits on how fast it can go. use this routine to throttle scl
'
_sclDelay
                call    #_sclDelayD2
                call    #_sclDelayD2
_sclDelay_ret   ret

_sclDelayD2
                mov     _delay, _delayCount
                djnz    _delay, #$
{
                nop
                nop
                nop
                nop
                nop
                nop
                nop
                nop
                nop
                nop
}
_sclDelayD2_ret ret

_delay          long    0
_delayCount     long    1

'---------------------------------------------------------------
' send 8 bits of data
' @param ack
' @param val
'
_sendVal
                mov     _ack,    #0              ' %00000000_00000000_00000000_11111111 ->
                ror     _val,    #8              ' %11111111_00000000_00000000_00000000
                andn    outa,    _scl            ' start with scl 0
                mov     _t0,      #8              ' Output val on sda
:loop
                shl     _val,    #1 wc           ' get next bit state into carry              
                muxc    outa,    _sda            ' set bit depending on carry
                call    #_sclDelay
                or      outa,    _scl            ' toggle clock
                call    #_sclDelay
                andn    outa,    _scl
                call    #_sclDelayD2
                djnz    _t0,      #:loop
:findack  
                andn    dira,    _sda            ' looking for ACK
                mov     _ack,    ina             ' get ack
                and     _ack,    _sda wz         ' keep only ack bit
                or      outa,    _scl            ' scl high
                call    #_sclDelay
                andn    outa,    _scl            ' scl low
                andn    outa,    _sda            ' sda low
                or      dira,    _sda            ' sda back to output
_sendVal_ret     ret


'---------------------------------------------------------------
' receive 8 bits of data
' @param ack
' @param val
'
_recVal
                mov     _val,    #0
                andn    dira,    _sda            ' sda input
                andn    outa,    _scl            ' clock starts at 0
                call    #_sclDelay
                mov     _t0,      #8              ' input 8 bits of val from sda
:loop
                or      outa,    _scl            ' toggle clock
                call    #_sclDelayD2
                shl     _val,    #1
                test    _sda,    ina wz          ' get bit in z
                muxnz   _val,    #1
                andn    outa,    _scl            ' toggle clock
                call    #_sclDelay
                djnz    _t0,      #:loop

:sendack
                or      dira,    _sda            ' set sda to out so device can interpret ACK
                or      outa,    _ack            ' ack is always on sda, but not always set
                call    #_sclDelayD2
                or      outa,    _scl       
                call    #_sclDelay

                andn    outa,    _scl            ' final clock pulse
                andn    outa,    _sda       
_recVal_ret      ret

'---------------------------------------------------------------
' constant vars, etc....
'
_zero           long    0
_bedtime        long    $1000

_bytes          res     1

eof             fit     $1ff

{{
=====================================================================
End of PASM Driver
=====================================================================
}}

{{
'------------------------------------------------------------------------------
TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
'------------------------------------------------------------------------------
}}
