/*
 * ==================================================
 * BbLCD.bas
 * This module is the BeanieBadge LCD driver.
 * It includes the BeanieBadge LCD PASM and access methods.
 * It also includes the BeanieBadge LCD 5x9 font.
 * ==================================================
 */
#include <stdio.h>
#include <propeller.h>
#include "UC1697a_LCD.h"

/*
 * ===================================================
 *  GameBabyLCD driver global variables
 */
HUB int LCD_bgcolor   = LCD_BLUE;
HUB int LCD_fgcolor   = LCD_WHITE;
HUB int LCD_row       = 0;
HUB int LCD_col       = 0;
HUB int LCD_contrast  = 0;
HUB int LCD_flag      = 0;

HUB int LCD_screen [LCD_screensize];

HUB int LCD_status    = 0;

struct _LCD_font_struct {
    unsigned int array;
    unsigned int table;
} volatile HUB LcdFont;

typedef struct _LCD_cmd_struct {
    unsigned int cmd;
    unsigned int address;
    unsigned int length;
    unsigned int parm[8];
} volatile HUB LCD_st;

LCD_st LcdCtrl;

/*
 * ===================================================
 * Start up the LCD ... WIP
 * Let the backlight to be controlled by the driver.
 */
void LCD_start(unsigned int *pasm)
{
    extern unsigned int binary_Fontms5x9_dat_start[];
    extern unsigned int binary_Font5x9Array_dat_start[];

    /* setup font */
    LcdFont.array  = (unsigned int) &binary_Font5x9Array_dat_start;
    LcdFont.table  = (unsigned int) &binary_Fontms5x9_dat_start;

    /* start LCD driver */
    LcdCtrl.cmd         = -1;
    LcdCtrl.address     = (unsigned int) &LcdFont;
    LcdCtrl.length      = 0;
    LcdCtrl.parm[0]     = 0;

    cognew(pasm,&LcdCtrl.cmd);

    /* wait for driver start */
    while(LcdCtrl.cmd)
        ; // wait for LCD_cmd

    /* initialize LCD */
    LCD_init();
    LCD_clear(LCD_WHITE);
}

/* ===================================================
 * initialize the LCD
 */
void LCD_init(void)
{
    /* ---------------------------
     *  wait for cmd clear
     */
    while(LcdCtrl.cmd)
        ; // wait for LcdCtrl.cmd

    /* ---------------------------
     *  power up routine from spec
     */
    LCD_writeCmd(LCD_SET_SystemReset);
    waitcnt(CLKFREQ/10+CNT);
    LCD_writeCmd(LCD_SET_TempComp | 0);
    LCD_writeCmd(LCD_SET_LcdMapCtrl | 6);       // x,y mirroring
    LCD_writeCmd(LCD_SET_RamAddrCtrl | 5);      // RowIncrDir, AutoInc Order, Wrap
    LCD_writeCmd(LCD_SET_LineRate | 3);
    LCD_writeCmd(LCD_SET_ColorMode | 2);        // 64K color mode
    LCD_writeCmd(LCD_SET_LcdBiasRatio | 3);
    LCD_writeCmd2(LCD_SET_VBiasPot,0x80);       // contrast control 0x8f max 0x60 min
    LCD_writeCmd(LCD_SET_DisplayEna | 3);       // enable display
    waitcnt(CLKFREQ/10+CNT);                        // wait before next write
    LCD_writeCmd(LCD_SET_AllPixelOn | 0);
    LCD_writeCmd(LCD_SET_InvDisplay | 0);       // don't invert display
    LCD_writeCmd(LCD_SET_ColorPattern | 1);     // normal pattern
}

/* ===================================================
 *  write register - a register has specific bits embedded in the address
 *  @param mreg  - register number with embedded operand bits
 */
void LCD_writeCmd(int mreg)
{
    LcdCtrl.address = mreg;                           // write register
    LcdCtrl.parm[0] = 0;
    LcdCtrl.cmd  = LCD_CMD_WRCMD;
    while(LcdCtrl.cmd)
        ;
}

/* ===================================================
 *  write register + an extra byte to device
 *  @param mreg  - register number with embedded operand bits
 *  @param mdata - second byte for command
 */
void LCD_writeCmd2(int mreg, int mdata)
{
    LcdCtrl.address = mreg;                           // write register + mdata
    LcdCtrl.parm[0] = mdata;
    LcdCtrl.cmd  = LCD_CMD_WRCMD2;
    while(LcdCtrl.cmd)
        ;
}

/* ===================================================
 *  write data to LCD graphics memory address
 *  @param madr - address to write
 *  @param mptr - buffer pointer
 *  @param mlen - num elements to write
 */
void LCD_writeData(int madr, int mptr, int mlen)
{
    LcdCtrl.address = madr;                           // write data to address
    LcdCtrl.parm[0] = mptr;
    LcdCtrl.length  = mlen;
    LcdCtrl.cmd  = LCD_CMD_WRDAT;
    while(LcdCtrl.cmd)
        ;
}

/* ===================================================
 *  read LCD graphics memory data
 *  @param madr - address to read
 *  @param mptr - buffer pointer
 *  @param mlen - num elements to read
 */
void LCD_readData(int madr, int mptr, int mlen)
{
    LcdCtrl.address = madr;                           // read data from address
    LcdCtrl.parm[0] = mptr;                           // read to mptr
    LcdCtrl.length  = mlen;
    LcdCtrl.cmd  = LCD_CMD_RDDATA;
    while(LcdCtrl.cmd)
        ;
}

/* ===================================================
 *  read LCD status
 */
int LCD_readStatus(void)
{
    LcdCtrl.address = 0;                              // read status register
    LcdCtrl.parm[0] = (int)&LCD_status;
    LcdCtrl.length  = 4;
    LcdCtrl.cmd  = LCD_CMD_RDSTAT;
    while(LcdCtrl.cmd)
        ;
    return LCD_status;
}

/* ===================================================
 *  clearScreen to a given color
 *  @param mc - pixel color
 */
void LCD_clear(int mc)
{
    LcdCtrl.parm[0]= mc;
    LcdCtrl.cmd    = LCD_CMD_CLRCOLOR;
    while(LcdCtrl.cmd)
        ;
}

/* ===================================================
 *  print character to display
 *  print ch with fcolor and bcolor to mx,my
 *  @param ch - char to put
 *  @param fcolor - char foreground color
 *  @param bcolor - char background color
 *  @param mx     - char left position
 *  @param my     - char top position
 */
void LCD_dispChar(char ch, int fcolor, int bcolor, int mx, int my)
{
    LcdCtrl.parm[0]  = (ch-0x20) % 95;
    LcdCtrl.parm[1]  = fcolor;
    LcdCtrl.parm[2]  = bcolor;
    LcdCtrl.parm[3]  = mx;
    LcdCtrl.parm[4]  = my;
    LcdCtrl.cmd      = LCD_CMD_SHOWCHR;
    while(LcdCtrl.cmd)
        ;
}

/* ===================================================
 *  print string to display
 * 
 *  draw string s with fgcolor and bgcolor to mx, my
 *  @param s - string to print/draw
 *  @param mx - left position
 *  @param my - top position
 */
void LCD_drawString(char *s, int mx, int my)
{
    int n = 0;
    int x = mx;
    char ch = s[n];
    while(ch)
    {
        LCD_dispChar(ch, LCD_fgcolor, LCD_bgcolor, x, my);
        x += LCD_FONTWIDTH;
        n ++;
        ch = s[n];
    }
}

/* ===================================================
 *  fill rectangle on display
 *  with mcolor to mx,my,width,height
 *  @param fcolor - foreground color
 *  @param mx - char left position
 *  @param my - char top position
 *  @param mw - width
 *  @param mh - height
 */
void LCD_fillRectangle(int fcolor, int mx, int my, int mw, int mh)
{
    LcdCtrl.parm[0]  = fcolor;
    LcdCtrl.parm[1]  = mx;
    LcdCtrl.parm[2]  = my;
    LcdCtrl.parm[3]  = mw;
    LcdCtrl.parm[4]  = mh;
    LcdCtrl.parm[5]  = 0;
    LcdCtrl.cmd      = LCD_CMD_BLOCK;
    while(LcdCtrl.cmd)
        ;
}

/* ===================================================
 *  draw rectangle on display
 *  with mcolor to mx,my,width,height
 *  @param fcolor - foreground color
 *  @param mx - char left position
 *  @param my - char top position
 *  @param mw - width
 *  @param mh - height
 *  @param mol- outline size
 */
void LCD_drawRectangle(int fcolor, int mx, int my, int mw, int mh, int mol)
{
    LcdCtrl.parm[0]  = fcolor;
    LcdCtrl.parm[1]  = mx;
    LcdCtrl.parm[2]  = my;
    LcdCtrl.parm[3]  = mw;
    LcdCtrl.parm[4]  = mh;
    LcdCtrl.parm[5]  = mol;
    LcdCtrl.cmd      = LCD_CMD_BLOCK;
    while(LcdCtrl.cmd)
        ;
}

/* ===================================================
 *  setPixel on display
 *  @param mc - pixel color
 *  @param mx - pixel left position
 *  @param my - pixel top position
 */
void LCD_setPixel(int mc, int mx, int my)
{
    LcdCtrl.parm[0]  = mc;
    LcdCtrl.parm[1]  = mx;
    LcdCtrl.parm[2]  = my;
    LcdCtrl.cmd      = LCD_CMD_SETPIXEL;
    while(LcdCtrl.cmd)
        ;
}

/* ===================================================
 *  print newline - scroll screen if necessary
 */
void LCD_newline(void)
{
    int x, y;
    int n = 0;
    LCD_col = 0;
    LCD_row++;
    if(LCD_row == LCD_rows)
    {
        LCD_row = 0;
        // copy size-c bytes from s+c to s
        //bytecopy(@LCD_screen, @LCD_screen[LCD_cols], LCD_lastrow)   //scroll lines
        for(n = 0; n < LCD_lastrow; n++)
            LCD_screen[n] = LCD_screen[n+LCD_cols];

        // clear the new last line
        //bytefill(@LCD_screen[LCD_lastrow], " ", LCD_cols)
        for(n = LCD_lastrow ; n < LCD_screensize; n++)
            LCD_screen[n] = ' ';

        // copy screen characters to device
        for(n = 0; n < LCD_lastrow; n++)
            LCD_putchar(LCD_screen[n]);

        for(n = 0; n < LCD_cols-1; n++) {
            x = LCD_col*LCD_FONTWIDTH+LCD_FONTOFFSET;
            y = LCD_row*LCD_FONTHEIGHT+LCD_FONTOFFSET;
            LCD_dispChar(' ', LCD_fgcolor, LCD_bgcolor, x, y);
            LCD_col++;
        }
        LCD_col = 0;
    }
}

/* ===================================================
 *  clear screen to current background color
 */
void LCD_cls(void)
{
    int n;
    LCD_row = 0;
    LCD_col = 0;
    for(n = 0; n < LCD_screensize; n++)
        LCD_screen[n] = ' ';
    LCD_clear(LCD_bgcolor);
}

/* ===================================================
 *  print character to screen at current position
 *  @param c - character to print
 */
void LCD_putchar(char c)
{
    int x, y;
    LCD_screen[LCD_row * LCD_cols + LCD_col] = c;
    if(c == 0xd) {
        LCD_newline();
    }
    else {
        x = LCD_col*LCD_FONTWIDTH+LCD_FONTOFFSET;
        y = LCD_row*LCD_FONTHEIGHT+LCD_FONTOFFSET;
        LCD_dispChar(c, LCD_fgcolor, LCD_bgcolor, x, y);
    }
    LCD_col++;
    if(LCD_col == LCD_cols)
        LCD_newline();
}

/* ===================================================
 *  set character x,y position
 *  @param mx - x or column position
 *  @param my - y or row position
 */
void LCD_setTextXY(int mx, int my)
{
  LCD_col = mx % LCD_cols;
  LCD_row = my % LCD_rows;
}

/* ===================================================
 *  get current x or column text position
 */
int LCD_getX(void)
{
    return LCD_col;
}

/* ===================================================
 *  get current y or row text position
 */
int LCD_getY(void)
{
    return LCD_row;
}

/* ===================================================
 *  set foreground color
 *  @param c - color
 */
void LCD_setFgColor(int c)
{
    LCD_fgcolor = c;
}

/* ===================================================
 *  get current foreground color
 */
int LCD_getFgColor(void)
{
    return LCD_fgcolor;
}
  
/* ===================================================
 *  set background color
 *  @param c - color
 */
void LCD_setBgColor(int c)
{
    LCD_bgcolor = c;
}

/* ===================================================
 *  get current background color
 */
int LCD_getBgColor(void)
{
    return LCD_bgcolor;
}

  
/* ===================================================
 *  print string to current x,y position
 *  @param s - string of chars LCD_print("Hello")
 */
void LCD_print(char *s)
{
    int n = 0;
    char ch = s[n];
    while(ch)
    {
        if(ch == 0xd)
            LCD_newline();
        else
            LCD_putchar(ch);
        n++;
        ch = s[n];
    }
}

/* ===================================================
 *  print zero terminated text to x,y position
 *  @param mx - x or column position
 *  @param my - y or row position
 *  @param s - string of chars LCD_printXY("Hello",5,10)
 */
void LCD_printXY(char *s, int mx, int my)
{
    int n = 0;
    char ch = s[n];
    LCD_col = mx % LCD_cols;
    LCD_row = my % LCD_rows;
    while(ch) {
        LCD_putchar(ch);
        n++;
        ch = s[n];
    }
}

/* ===================================================
 *  Print a zero-terminated string to current text position
 *  @param s - string of chars LCD_str("Hello")
 */
void LCD_str(char *s)
{
    int n = 0;
    char ch = s[n];
    while(ch) {
        LCD_putchar(ch);
        n++;
        ch = s[n];
    }
}

/* ===================================================
 *  Print a decimal number to current text position
 *  @param value - decimal number
 */
void LCD_dec(int value)
{
    int n = value;
    int length = 10;
    int result = 0;

    if(value < 0) {
        value = 0-value;
        LCD_putchar('-');
    }

    n = 1000000000;

    while (length > 0) {
        length--;
        if(value >= n) {
            LCD_putchar(value / n + '0');
            value %= n;
            result++;
        }
        else if(result) {
            LCD_putchar('0');
        }
        else if(n == 1) {
            LCD_putchar('0');
        }
        n = n / 10;
    }
}

static int LCD_hexarray[16] = {
    '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
};

/* ===================================================
 *  Print a hexidecimal number to current text position
 *  @param value - decimal number
 *  @param digits - number of digits to print
 */
void LCD_hex(int value, int digits)
{
    int n = 0;
    while(digits > 0) {
        digits--;
        n = (value >> (digits<<2)) & 0xf;
        LCD_putchar(LCD_hexarray[n]);
    }
}

/* ===================================================
 *  Print a binary number to current text position
 *  @param value - number
 *  @param digits - number of digits to print
 */
void LCD_bin(int value, int digits)
{
    char dig = 0;
    int n = digits-1;
    while(n > -1) {
        dig = '0' + ((value & (1<<n)) >> n);
        LCD_out(dig);
        n--;
    }
}

/* ===================================================
 *  Output a character
 * 
 *  @param c - interpreted as follows
 * 
 *   0x00 = clear screen
 *   0x01 = home
 *   0x08 = backspace
 *   0x09 = tab (8 spaces per)
 *   0x0A = set X position (X follows)
 *   0x0B = set Y position (Y follows)
 *   0x0C = set color (color follows)
 *   0x0D = return
 *   others = printable characters
 * ===================================================
 */
void LCD_out(char c)
{
    if(LCD_flag == 0) {
        if(c == 0) {
            LCD_cls();
        }
        else if(c == 1) {
            LCD_col = 0;
            LCD_row = 0;
        }
        else if(c == 8) {
            if(LCD_col) {
                LCD_col--;
            }
        }
        else if(c == 9) {
            while(LCD_col & 7) {
                  LCD_putchar(' ');
            }
        }
        else if(c == 0xa) {
            LCD_flag = c;
            return;
        }
        else if(c == 0xb) {
            LCD_flag = c;
            return;
        }
        else if(c == 0xc) {
            LCD_flag = c;
            return;
        }
        else {
            LCD_putchar(c);
        }
    }
    else if(LCD_flag == 0xA) {
        LCD_col = c % LCD_cols;
    }
    else if(LCD_flag == 0xB) {
        LCD_row = c % LCD_rows;
    }
    else if(LCD_flag == 0xC) {
        LCD_fgcolor = c;
    }
    LCD_flag = 0;
}
